/* $Id: init.c,v 1.5 1999/06/08 00:38:20 marcus Exp $
******************************************************************************

   Bse extension initialization.

   Copyright (C) 1999 Andreas Beck	[becka@ggi-project.org]
		 2001 Christoph Egger	[Christoph_Egger@t-online.de]
  
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <string.h>

#include <ggi/internal/bse.h>


#include <ggi/ovl.h>
#include <ggi/blt.h>


/* Static variables 
 */

static int _ggibseLibIsUp = 0;

struct ressourcelist_t *resslist = NULL;



int ggiBseInit(void)
{
	int err = 0;

	_ggibseLibIsUp++;
	if (_ggibseLibIsUp > 1) {
		if (_ggibseLibIsUp >= BSE_MAX_USAGE) {
			/* Max initializes reached! */
			_ggibseLibIsUp--;
			return -1;
		}	/* if */

		return 0;	/* Initialize only at first call. */
	}	/* if */


	if (err = ggiOvlInit()) != 0) {
		fprintf(stderr,"LibBSE: couldn't Initilize LibOVL.\n");
		_ggibseLibIsUp--;
		return err;
	}	/* if */

	if (err = ggiBltInit()) != 0) {
		fprintf(stderr,"LibBSE: couldn't Initilize LibBLT.\n");
		_ggibseLibIsUp--;
		return err;
	}	/* if */


	return 0;
}	/* ggiBseInit */


int ggiBseExit(void)
{
	int rc = 0;
    
	if (!_ggibseLibIsUp) {
		return -1;
	}	/* if */

	if (_ggibseLibIsUp > 1) {
		/* Exit only at last call */
		_ggibseLibIsUp--;
		return 0;
	}	/* if */
	

#warning  TODO: Check the return-code here
	rc = ggiBltExit();
	rc = ggiOvlExit();


	_ggibseLibIsUp = 0;

	return rc;
}	/* ggiBseExit */


int ggiBseAttach(ggi_visual_t vis)
{
	int rc = 0;
	struct ressourcelist_t *new_res = NULL;
	struct ressourcelist_t *cur_res = NULL;

	new_res = calloc(1, sizeof(struct ressourcelist_t));
	if (!new_res) {
		return GGI_ENOMEM;
	}	/* if */

#warning  TODO: Check the return-code here
	rc = ggiOvlAttach(vis);
	rc = ggiBltAttach(vis);

	if (rc != 0) {
		return rc;
	}	/* if */

	new_res->vis = vis;
	new_res->first = NULL;

	ADD_NEW_RES(cur_res, new_res);

	return rc;
}	/* ggiBseAttach */


int ggiBseDetach(ggi_visual_t vis)
{
	int rc = 0;
	struct ressourcelist_t *cur_res = NULL;

	ggibse_t next;
	ggibse_t remains;


	FIND_RESSOURCE(vis, cur_res);
	if (cur_res == NULL) {
		fprintf(stderr, "Sorry! But this visual was never attached before!\n");
		return -1;
	}	/* if */

	remains = cur_res->first;
	while (remains != NULL) {
		next = remains->next;
		ggiBseFree(remains);
		remains = next;
	}	/* while */


#warning  TODO: Check the return-code here
	rc = ggiBltDetach(vis);
	rc = ggiOvlDetach(vis);


	DESTROY_RESSOURCE(cur_res);


	return rc;
}	/* ggiBseDetach */
