/* $Id: stubs.c,v 1.2 1999/01/12 20:06:30 becka Exp $
******************************************************************************

   Bse extension stubs.

   Copyright (C) 1999 Andreas Beck	[becka@ggi-project.org]
		 2001 Christoph Egger	[Christoph_Egger@t-online.de]
  
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>

#include <ggi/internal/bse.h>


/* Allocating/Destroying a BSE.
 */
ggibse_t ggiBseAllocate(ggi_visual_t vis, int width, int height,
			enum bsetype type)
{
	ggibse_t newsprite;
	ggibse_t bse_index;

	struct ressourcelist_t *cur_res = NULL;

	FIND_RESSOURCE(vis, cur_res);
	if (cur_res == NULL) {
		fprintf(stderr, "Sorry! This visual was never attached before!\n");
		return NULL;
	}	/* if */


	switch (type) {
	case BSE_SPRITE:
		newsprite = ggiOvlAllocate(vis, width, height);
		if (!newsprite) {
			/* fall back to emusprite */
			type = BSE_EMUSPRITE;
			goto emusprite;
		}	/* if */
		break;

	case BSE_BOB:
	case BSE_EMUSPRITE:
emusprite:
		newsprite = malloc(sizeof(ggibse_t));
		if (!newsprite) {
			return NULL;
		}	/* if */

		newsprite->vis = vis;
		newsprite->x   =
		newsprite->y   = 0;
		newsprite->w   = width;
		newsprite->h   = height;

		switch (type) {
		case BSE_BOB:
			/* BOBs have no notion of being hidden */
			newsprite->hidecount = 0; 
			break;

		default:
			/* come up hidden ! */
			newsprite->hidecount = 1;
			break;
		}	/* switch */
		newsprite->type = type;
		break;

	default:
		fprintf(stderr, "Unsupported type of sprite: %i!\n", type);
		return NULL;

	}	/* switch */


	switch (type) {
	case BSE_BOB:
		newsprite->subtype.bob = ggiBltAllocate(vis, width, height);
		if (!newsprite->subtype.bob) {
			return NULL;
		}	/* if */
		break;

	case BSE_SPRITE:
		/* Nothing todo here */
		break;

	case BSE_EMUSPRITE:
		newsprite->subtype.es.bob = ggiBltAllocate(vis, width, height);
		if (!newsprite->subtype.es.bob) {
			return NULL;
		}	/* if */
		newsprite->subtype.es.backgr = NULL;
		break;

	default:
		fprintf(stderr, "Unsupported type of sprite: %i!\n", type);
		return NULL;

	}	/* switch */


	ADD_NEW_BSE(bse_index, cur_res->first, newsprite);

	return newsprite;
}	/* ggiBseAllocate */



int ggiBseFree(ggibse_t bob)
{
	int rc = 0;
	struct ressourcelist_t *cur_res = NULL;


	if (!bob) {
		/* Sprite wasn't alloced. Keep it simple. */
		return 0;
	}	/* if */


	FIND_RESSOURCE(bob->vis, cur_res);
	if (cur_res == NULL) {
		fprintf(stderr, "Sorry! This visual was never attached before!\n");
		return NULL;
	}	/* if */


	switch (bob->type) {
	case BSE_BOB:
		rc = ggiBltFree(bob->subtype.bob);
		break;

	case BSE_SPRITE:
		rc = ggiOvlFree(bob);
		break;

	case BSE_EMUSPRITE:
		if (bob->subtype.es.backgr) {
			free(bob->subtype.es.backgr);
		}	/* if */
		rc = ggiBltFree(bob->subtype.es.bob);
		break;

	default:
		fprintf(stderr, "Unsupported type of sprite: %i!\n", type);
		break;

	}	/* switch */


	DESTROY_BSE(bob, cur_res->first);

	return rc;
}




/* Hiding and showing
 */
void ggiBseHideAboveCollide(ggibse_t bob,int x,int y) 
{
	ggibse_t curr = bob;

	while ( (curr = curr->prev) ) {
		if (curr->type == BSE_EMUSPRITE) {
			if (curr->x + curr->w <= x ||
			    curr->y + curr->h <= y ||
				  x + bob->w <=curr->x ||
				  y + bob->h <=curr->y )
				continue; 
			ggiBseHideAboveCollide(curr,curr->x,curr->y);
			ggiBseHide(curr);
		}	/* if */
	}	/* while */
}	/* ggiBseHideAboveCollide */


void ggiBseShowAboveCollide(ggibse_t bob,int x,int y) 
{
	ggibse_t curr = bob;

	while ( (curr = curr->prev) ) {
		if (curr->type == BSE_EMUSPRITE) {
			if (curr->x + curr->w <= x ||
			    curr->y + curr->h <= y ||
				  x + bob->w <= curr->x ||
				  y + bob->h <= curr->y )
				continue; 
			ggiBseShow(curr);
			ggiBseShowAboveCollide(curr,curr->x,curr->y);
		}	/* if */
	}	/* while */
}	/* ggiBseShowAboveCollide */



/* Populate the sprite
 */
int ggiBseSetImage(ggibse_t bob,ggi_visual_t vis,int x,int y,
		 unsigned char *alpha)
{
	int rc = 0;

	switch(bob->type) {
	case BSE_BOB:
		rc = ggiBltSetImage(bob->subtype.bob, vis, x, y, alpha);
		return rc;

	case BSE_SPRITE:
		rc = ggiOvlSetImage(bob, vis, x, y, alpha);
		fprintf(stderr, "Oops! - This failed! Switch to EMU-Sprites!\n");
		return rc;

	case BSE_EMUSPRITE:
		if (bob->subtype.es.backgr == NULL) {
			bob->subtype.es.backgr =
				malloc(bob->w * bob->h * sizeof(ggi_pixel));
			if (!bob->subtype.es.backgr) {
				return GGI_ENOMEM;
			}	/* if */
		}	/* if */

		rc = ggiBltSetImage(bob->subtype.es.bob,vis,x,y,alpha);
		return rc;

	default:
		fprintf(stderr, "Unsupported type of sprite: %i!\n", type);
		return -1;
	}	/* switch */

	return rc;
}	/* ggiBseSetImage */


/* Move the sprite around.
 */
int ggiBseMove(ggibse_t bob,int x,int y)
{
	int rc;
	
	if (x == bob->x && y == bob->y) {
		return 0;	/* It's already there ! */
	}	/* if */

	if (bob->hidecount) {
		bob->x=x;
		bob->y=y;
		return 0;	/* It's hidden. */
	}	/* if */

	switch(bob->type) {
	case BSE_BOB:
		rc = ggiBltBlit(bob->subtype.bob, x, y);
		break;

	case BSE_SPRITE:
		rc = ggiOvlMove(bob, x, y);
		fprintf(stderr, "Oops! - This failed! Switch to EMU-Sprites!\n");
		return rc;

	case BSE_EMUSPRITE:
		{
		int oldx = bob->x;
		int oldy = bob->y;

		ggiBseHideAboveCollide(bob, oldx, oldy);
		ggiBseHideAboveCollide(bob, x, y);
		ggiPutBox(bob->vis, bob->x,bob->y,
				bob->w,bob->h, bob->subtype.es.backgr);
		ggiGetBox(bob->vis, x,y, bob->w,bob->h,bob->subtype.es.backgr);
		rc = ggiBltBlit(bob->subtype.es.bob, x, y);
		ggiBseShowAboveCollide(bob, x, y);
		ggiBseShowAboveCollide(bob, oldx, oldy);
		}
		break;

	default:
		fprintf(stderr, "Unsupported type of sprite: %i!\n", type);
		return -1;
	}	/* switch */

	bob->x = x;
	bob->y = y;

	return rc;
}	/* ggiBseMove */


int ggiBseHide(ggibse_t bob)
{
	int rc;

	if (bob->type == BSE_BOB) {	
		return GGI_EARGINVAL;
	}	/* if */

	if (bob->hidecount++) {
		return 0;	/* It's already hidden. */
	}	/* if */

	switch(bob->type) {
	case BSE_BOB:
		/* A BOB cannot be hidden. */
		return GGI_EARGINVAL;

	case BSE_SPRITE:
		/* Hide it. */
		rc = ggiOvlShow(bob, !bob->hidecount);
		fprintf(stderr, "Oops! - This failed! Switch to EMU-Sprites!\n");
		return rc;

	case BSE_EMUSPRITE:
		ggiBseHideAboveCollide(bob,bob->x,bob->y);
		ggiPutBox(bob->vis,bob->x,bob->y,
				bob->w,bob->h,bob->subtype.es.backgr);
		rc=0;
		ggiBseShowAboveCollide(bob,bob->x,bob->y);
		break;

	default:
		fprintf(stderr, "Unsupported type of sprite: %i!\n", type);
		return -1;
	}	/* switch */

	return rc;
}	/* ggiBseHide */


int ggiBseShow(ggibse_t bob)
{
	int rc = 0;
	
	if (bob->type == BSE_BOB) {
		/* A BOB cannot be hidden. */
		return GGI_EARGINVAL;
	}	/* if */

	if (bob->hidecount <= 0) {
		return GGI_EARGINVAL;	/* Too often ! */
	}	/* if */

	if (--(bob->hidecount) != 0) {
		return 0;	/* We are still hidden. */
	}	/* if */

	switch(bob->type) {
	case BSE_BOB:
		/* A BOB cannot be hidden. */
		return GGI_EARGINVAL;

	case BSE_SPRITE:
		/* Hide it. */
		rc = ggiOvlShow(bob, !bob->hidecount);
		fprintf(stderr, "Oops! - This failed! Switch to EMU-Sprites!\n");
		return rc;

	case BSE_EMUSPRITE:
		ggiBseHideAboveCollide(bob,bob->x,bob->y);

		/* Save box */
		ggiGetBox(bob->vis, bob->x,bob->y,
				bob->w,bob->h,bob->subtype.es.backgr);
		rc = ggiBltBlit(bob->subtype.es.bob,bob->x,bob->y);
		ggiBseShowAboveCollide(bob,bob->x,bob->y);
		break;

	default:
		fprintf(stderr, "Unsupported type of sprite: %i!\n", type);
		return -1;
	}	/* switch */

	return rc;
}	/* ggiBseShow */


/* These are for EMU_SPRITEs. You can use these for hiding them for drawing.
 */
int ggiBseSoftHide	(ggibse_t bob)
{
	switch(bob->type) {
	case BSE_BOB:
	default:
		return GGI_EARGINVAL;
	case BSE_SPRITE:
		return 0;	/* NOP for real sprites */
	case BSE_EMUSPRITE:
		return ggiBseHide(bob);
	}	/* switch */
}	/* ggiBseSoftHide */


int ggiBseSoftShow	(ggibse_t bob)
{
	switch(bob->type) {
	case BSE_BOB:
	default:
		return GGI_EARGINVAL;
	case BSE_SPRITE:
		return 0;	/* NOP for real sprites */
	case BSE_EMUSPRITE:
		return ggiBseShow(bob);
	}	/* switch */
}	/* ggiBseSoftShow */


/* Helper functions for Alpha channels
 */
 
unsigned char *ggiBseMakeAlphaFromPixelkey(ggibse_t bob,
	ggi_visual_t vis,int x,int y,ggi_pixel pix) {

	unsigned char *rc,*alph;
	ggi_pixel gp;
	int xx,yy;

	if ((rc = malloc(bob->w*bob->h*sizeof(unsigned char)))) {
		alph = rc;
		for (yy = 0;yy < bob->h; yy++) {
			for (xx = 0;xx < bob->w; xx++) {
				ggiGetPixel(vis, x + xx, y + yy, &gp);
				*alph++ = (gp == pix) ? 0xff : 0x00;
			}	/* for */
		}	/* for */
	}	/* if */
	return rc;
}	/* ggiBseMakeAlphaFromPixelkey */


unsigned char *ggiBseMakeAlphaFromColorkey(ggibse_t bob,
	ggi_visual_t vis,int x,int y,ggi_color *col)
{
	return ggiBseMakeAlphaFromPixelkey(bob,vis,x,y,ggiMapColor(vis,col));
}	/* ggiBseMakeAlphaFromColorkey */


static unsigned char func_treshold(unsigned char inval,void *tresh)
{
	return inval > *((unsigned char *)tresh) ? 0xff : 0x00;
}	/* func_treshold */


static unsigned char func_invert(unsigned char inval,void *unused)
{
	return 0xff - inval;
}	/* func_invert */


unsigned char *ggiBseChangeAlphaByFunction(ggibse_t bob, 
	unsigned char *alpha,
	unsigned char (*func)(unsigned char in,void *parms),void *parms)
{

	int xx,yy;

	for (yy = 0; yy < bob->h; yy++) {
		for (xx = 0; xx < bob->w; xx++) {
			*alpha = func(*alpha,parms);
		}	/* for */
	}	/* for */

	return alpha;
}	/* ggiBseChangeAlphaByFunction */


unsigned char *ggiBseChangeAlphaTreshold(ggibse_t bob, 
				unsigned char *alpha,
				unsigned char treshold)
{
	return ggiBseChangeAlphaByFunction(bob,alpha,func_treshold,&treshold);
}	/* ggiBseChangeAlphaTreshold */


unsigned char *ggiBseChangeAlphaInvert(ggibse_t bob, 
				unsigned char *alpha)
{
	return ggiBseChangeAlphaByFunction(bob,alpha,func_invert,NULL);
}	/* ggiBseChangeAlphaInvert */
